/*
 * rgSDK/rg3DRom/rg3DRom.h
 *
 * Copyright (C) 2004 Amour Hassan
 */

#include "rg3DRom.h"
#include "rgPort.h"
#include <iostream>
#include <string.h>
#include <math.h>

#define BONE_DEBUG

sk3DRomBone::sk3DRomBone()
{
	bname[0] = '\0';
	bname[ SK3DROM_BONE_NAME_LEN - 1] = '\0';
	pj = 0;
	mid = -1;
}

rgVector sk3DRomBone::get_pivot_joint() const
{
	if( pj <= 0) return a;
	return b;
}

rgVector sk3DRomBone::get_rotate_joint() const
{
	if( pj <= 0) return b;
	return a;
}



//void sk3DRomSkelFrame::update_skel_fcs( sk3DRomDrawContext &dc) const
//{
//	
//}

sk3DRomSkel::sk3DRomSkel()
{
}


void hlp_update_fcs( sk3DRomDrawContext &dc, const sk3DRomBone &b,
	const rgCoordSystem mcs, const sk3DRomSkel &sk,
	const sk3DRomSkelFrame &sf)
{
	{ // Generate the "fcs" for all immediate slaves.
		for( int i = 0; i < int( b.link.size()); i++) {
			dc.set_fcs( (b.link[i]).bid,
				mcs.map_csys( sf.b[ (b.link[i]).bid].rcs));
		}
	}

	{ // Produce the "fcs" for the slaves of slaves.
		for( int i = 0; i < int( b.link.size()); i++) {
			hlp_update_fcs( dc, sk.b[ (b.link[i]).bid],
				dc.get_fcs( (b.link[i]).bid), sk, sf);
		}
	}
}

void sk3DRomSkel::update_fcs( sk3DRomDrawContext &dc,
	const sk3DRomSkelFrame &sf) const
{
	// Generate the "fcs" for all master bones.
	// FIXME: Master bones may be mounted, so for now, get_full_rcs is
	//	used. This is inefficient, use an efficient method.
	{
		for( int i = 0; i < int( rb.size()); i++)
			dc.set_fcs( rb[i], get_full_bone_rcs( rb[i], sf));
	}

	{ // Generate the "fcs" for all slaves.
		for( int i = 0; i < int( rb.size()); i++) {
			hlp_update_fcs( dc, b[ rb[i]],
				dc.get_fcs( rb[i]), *this, sf);
		}
	}
}

void sk3DRomSkel::translate_rcs( const int bid, sk3DRomSkelFrame &p_sf,
        const rgVector v) const
{
	rgCoordSystem mcs;
	get_full_bone_rcs( bid, p_sf, mcs);
	p_sf.b[bid].rcs.p = p_sf.b[bid].rcs.p + mcs.change_vector_csys( v);
}

void sk3DRomModel::swap_pivot_joint( const int bid)
{
	if( (bid < 0) || (bid >= int( skel.b.size()))) return;
	if( 0 == skel.b[bid].pj) set_pivot_joint( bid, 1);
	else set_pivot_joint( bid, 0);
}

// Description: Builds a list of root bone indicies. All bones that have no
//      master bone are root bones.
void sk3DRomSkel::build_root_bone_list()
{
	rb.clear();
	const short istop = short( b.size());
	for( short i = 0; i < istop; i++)
		if( (b[i]).mid < 0) rb.push_back( i);
}

// Description: Initializes "src" to contain the index values of all root
//	bones. "mmb" will hold the "root master mount bone" for each bone, or
//	-1 if a particular bone has no Master Mount Bone (mmb). The "root
//	master mount bone (rmmb)" is the root bone of mmb (rmmb might just be
//	mmb itself if mmb is not bound to another bone). In simpler terms,
//	mmb will contain either -1 or a root bone (the root bone can either be
//	the bones mount master or the bones root mount master).
inline void hlp_init_root_bone_mount_order_list( int *src, int *mmb, int n,
	const sk3DRomSkel &s, const sk3DRomSkelFrame &sf)
{
	while( --n) {
		int i = s.rb[i];
		src[n] = i;
		int k = mmb[i] = sf.b[i].mmb;
		if( (k >= 0) && !(s.b[k].mlist.empty())) {
			mmb[i] = s.b[k].mlist[ s.b[k].mlist.size() - 1];
		}
	}
}

inline int hlp_gen_dependency_list( int i, int *dep, const int *chk,
	const int *mmb)
{
	int n = 0;
	while( (i >= 0) && !(chk[i])) {
		dep[ n++] = i;
		i = mmb[i];
	}
	return n;
}

inline void hlp_write_out_dependency_list( const int *dep, int *chk, int n,
	std::vector <int> &mlist)
{
	while( (n--) > 0) {
		mlist.push_back( dep[n]);
		chk[ dep[n]] = 1;
	}
}

inline void hlp_put_in_mount_order( int *src, int *mmb, int *chk, int *dep,
	const int n, const sk3DRomSkel &s, const sk3DRomSkelFrame &sf,
	std::vector <int> &mlist)
{
	for( int i = 0; i < n; i++) {
		hlp_write_out_dependency_list( dep, chk,
			hlp_gen_dependency_list( src[i], dep, chk, mmb),
			mlist);
	}
}

// TODO: This function MUST be tested so that it may be used to optimize
//	the computation of skeleton frames making use of mounted bones.
void sk3DRomSkel::gen_mount_order_root_bone_list( std::vector <int> &mlist,
	const sk3DRomSkelFrame &p_sf)
{
	mlist.clear();
	if( rb.empty()) return;

	const int n = int( rb.size());
	const int buf_size = n * 4;
	int *src = new int[ buf_size];
	int *mmb = src + n;
	int *chk = src + (n * 2);
	int *dep = src + (n * 3);
	hlp_init_root_bone_mount_order_list( src, mmb, n, *this, p_sf);
	hlp_put_in_mount_order( src, mmb, chk, dep, n, *this, p_sf, mlist);
	delete [] src;
}

void sk3DRomSkel::remove_root_bone( const int bid)
{
	if( (bid < 0) || (bid >= int( b.size()))) return;
	for( int i = 0; i < int( rb.size()); i++) {
		if( bid == rb[i]) {
			rb.erase( rb.begin() + i);
			return;
		}
	}
}

// Description: Adds a given bone to the root bone list. If the given bone
//	is already a part of the root bone list, then nothing is done.
void sk3DRomSkel::add_root_bone( const int bid)
{
	if( (bid < 0) || (bid >= int( b.size()))) return;
	for( int i = 0; i < int( rb.size()); i++)
		if( bid == rb[i]) return;
	rb.push_back( bid);	
}

/*
void sk3DRomSkel::add_bone_link( sk3DRomBone &p_b, const int link)
{
	const int istop = int( p_b.link.size());
	for( int i = 0; i < istop; i++)
		if( link == p_b.link[i].bid) return;
	p_b.link.push_back( sk3DRomBoneLink( link, ));
}
*/

void sk3DRomSkel::remove_bone_link( sk3DRomBone &p_b, const int link)
{
	const int istop = int( p_b.link.size());
	for( int i = 0; i < istop; i++) {
		if( link == p_b.link[i].bid) {
			p_b.link.erase( p_b.link.begin() + i);
			return;
		}
	}
}

void sk3DRomSkel::notify_sequences_of_frame_delete( const int fid)
{
	for( int i = 0; i < sequence_count(); i++) {
		// FIXME: An animation frame(s) is being deleted. Entities
		//	referencing this or other animation frames
		//	must be notified.
		(seq[i]).delete_all_frames_with_fid( fid);
	}
}

void sk3DRomSkel::notify_sequences_of_frame_swap( const int old_fid,
	const int new_fid)
{
	for( int i = 0; i < sequence_count(); i++) {
		for( int j = 0; j < int( seq[i].frame.size()); j++) {
			if( seq[i].frame[j].fid() == old_fid)
				seq[i].frame[j].set_fid( new_fid);
		}
	}
}

// Parameters:
//	fid - The index number of the skeleton frame in the "sf" array.
void sk3DRomSkel::delete_skeleton_frame( const int fid)
{
	if( (fid < 0) || (fid >= frame_count())) return;
	notify_sequences_of_frame_delete( fid);
	const int last = frame_count() - 1;
	if( fid != last) {
		notify_sequences_of_frame_swap( last, fid);
		sf[fid] = sf[last];
	}
	sf.erase( sf.begin() + last);
}

void sk3DRomModel::delete_all_skeleton_frames()
{
	while( skel.frame_count() > 0)
		skel.delete_skeleton_frame( skel.frame_count() - 1);
}

// Description: Notifies all verts of the bones they are linked to.
void sk3DRomModel::set_bid_for_all_verts()
{
	clear_bid_for_all_verts();
	const int istop = int( skel.b.size());
	for( int i = 0; i < istop; i++) {
		{
			sk3DRomBone &b = skel.b[i];
			const int jstop = int( b.vid.size());
			for( int j = 0; j < jstop; j++) v[ b.vid[j]].m_bid = i;
		}
	}
}

void sk3DRomModel::build_bone_less_list()
{
	bone_less_verts.clear();
	if( int( v.size()) <= 0) return;

	char *data = new char[ v.size()];
	memset( data, 0, v.size());

	// Step through all bones and mark off owned verticies.
	const int istop = int( skel.b.size());
	int i;
	for( i = 0; i < istop; i++) {
		{
			const sk3DRomBone &b = skel.b[i];
			const int jstop = int( b.vid.size());
			for( int j = 0; j < jstop; j++) data[ b.vid[j]] = 1;
		}
	}

	// Add unowned verticies to the bone-less list.
	for( i = 0; i < int( v.size()); i++)
		if( !(data[i])) bone_less_verts.push_back( i);

	delete [] data;
}

// Description: Creates a Joint Coordinate System (jcs) for every bone in
//	the skeleton. This is required for "MODEC" (or older) 3DRom skeletons
//	(since they did not support jcs's).
void sk3DRomModel::auto_gen_joints()
{
        const short istop = short( skel.b.size());
        for( short i = 0; i < istop; i++) skel.create_bone_jcs( skel.b[i]);
}

// Description: Creates a Joint Coordinate System (jcs) for the given bone.
//	This step is required for bones loaded from "MODEC" 3DRom files or
//	lower (older) and also required for newly created bones. All bones
//	must have valid jcs's before they can be used for animations and
//	such in skeletons. It is valid to recreate the jcs of a given bone
//	(calling "create_bone_jcs" multiple times on a single bone), though
//	this is unnessasary.
//
void sk3DRomSkel::create_bone_jcs( sk3DRomBone &b)
{
	// Set the "bone" vector
	b.jcs.p = b.get_pivot_joint();
	b.jcs.up = (sk3DRom_get_bone_vector( b)).normalize();

	// Determine the "best" normal for the current bone.
	rgVector tmp_x( 1, 0, 0);
	rgVector tmp_y( 0, 1, 0);
	const float k = b.jcs.up.dot( tmp_x);
	const float t = b.jcs.up.dot( tmp_y);

	// NOTE: The "best" normal is the one that is the
	// most orthogonal (the one that produces a dot
	// product with "up" that is closest to zero).
	rgVector n( tmp_y);
	if( fabs( k) < fabs( t)) n = tmp_x;

	// Make "n" be orthogonal to the bone.
	n = n - (b.jcs.up.dot( n) * b.jcs.up);
	n.normalize();

	b.jcs.eye = n;
	b.jcs.side = (b.jcs.up).cross( n);
}

// Description: Determines the "rcs" for a given bone. If an invalid bone is
//	given, then the standard coordinate system is returned.
//
//	If the given bone is "mounted", then it is neccasarily a root bone.
//	If the bone is "mounted", then the rcs of the Master Mount Bone (mmb)
//	must be calculated first.
//
//	If the bone is not mounted, and has a list of master bones, then the
//	the last/primary master bone might be mounted. Therefore, all bones
//	with master lists must check the last/primary master to see if it is
//	mounted.
rgCoordSystem sk3DRomSkel::get_full_bone_rcs( const int bid,
	const sk3DRomSkelFrame &s, rgCoordSystem &mcs) const
{
	//if( bid < 0) *((int *)0) = 0;
	if( bid < 0) return rgCoordSystem();

	if( s.b[bid].mmb >= 0) {
		mcs = get_full_bone_rcs( s.b[bid].mmb, s);

	} else if( !(b[bid].mlist.empty())) {
		const short istart = short( b[bid].mlist.size() - 1);
		if( s.b[ b[bid].mlist[istart]].mmb >= 0) {
			mcs = get_full_bone_rcs(
				s.b[ b[bid].mlist[istart]].mmb, s);
		}
		for( short i = istart; i >= 0; i--)
			mcs = mcs.map_csys( s.b[ b[bid].mlist[i]].rcs);
	}

	// Apply the bone's rcs.
	return mcs.map_csys( s.b[bid].rcs);
}

int sk3DRomSkel::get_root_master( const int bid, const sk3DRomSkelFrame &s)
	const
{
	if( (bid < 0) || (bid >= int( b.size()))) return -1;
	if( s.b[bid].mmb >= 0) return get_root_master( s.b[bid].mmb, s);
	else if( !(b[bid].mlist.empty())) {
		const int mid = b[bid].mlist[ b[bid].mlist.size() - 1];
		if( s.b[ mid].mmb >= 0)
			return get_root_master( s.b[ mid].mmb, s);
		return mid;
	}
	return bid;
}

// Description: "Binds" bone verticies to the bone's Joint Coordinate System
//      (jcs).
//      Binding of the skeleton must start at the root bones. Then, from
//      the root bones, binding continues all the way up to the last slave
//      bones.
//      First, a list of root bones must be built, once done, then the
//      binding will start at the root bones.
void sk3DRomSkel::bind_skeleton()
{
	build_root_bone_list();

	// Start the binding at the root bones.
	const short istop = short( rb.size());
	for( short i = 0; i < istop; i++) {
		// Both root bone joints are controlled by the root bone's
		// own joint basis.
		//b[ rb[i]].ja = b[ rb[i]].jcs.change_point_csys( b[ rb[i]].a);
		//b[ rb[i]].jb = b[ rb[i]].jcs.change_point_csys( b[ rb[i]].b);
		bind_bone_to_jcs( 0, b[ rb[i]]);
		bind_slaves( rb[i]);
	}

	reset_all_skel_frames();
}

void sk3DRomSkel::bind_slaves( const short mid)
{
	// Bind all of this master's slaves.
	const short istop = short( b[mid].link.size());
	short i;
	rgCoordSystem mcs = get_full_bone_jcs( mid);
	for( i = 0; i < istop; i++) {

		{
			// Bind the bone's joints to the jcs, then bind the
			// jcs to the master bone.
			//short j = b[mid].link[i].bid;

			bind_bone_to_jcs( &mcs, b[ b[mid].link[i].bid]);
			//b[j].ja = b[j].jcs.change_point_csys( b[j].a);
			//b[j].jb = b[j].jcs.change_point_csys( b[j].b);

			// The jcs (Joint Coordinate System) of the slave is
			// relative to the master. Currently, the slave's jcs
			// is relative to the standard coord system. Therefore,
			// the slave's jcs must be converted.

			//b[j].jcs = mcs.change_csys_csys( b[j].jcs);
		}
	}

	// Bind the slave's slave bones
	for( i = 0; i < istop; i++)
		bind_slaves( b[mid].link[i].bid);
}

void sk3DRomSkel::bind_bone_to_bone( const int mid, const int bid)
{
	if( bone_id_is_invalid( bid)) return;
	if( (mid >= int( b.size()))) return;
	if( bid == mid) return;
	unbind_bone( bid);
	sk3DRomBone &bn = b[bid];
	if( (bn.mid = mid) >= 0) {
		if( bid == b[mid].mid) unbind_bone( mid);
		const int istop = int( b[mid].link.size());
		for( int i = 0; i < istop; i++)
			if( bid == (b[mid].link[i].bid)) return;
		rgCoordSystem jcs( get_full_bone_jcs( bn.mid));
		bind_bone_to_jcs( &jcs, bn);
		remove_root_bone( bid);
		sk3DRomBoneLink link;
		link.bid = bid;
		link.jid = 0;
		b[mid].link.push_back( link);
	}

	build_master_list();
	skout.b[bid].rcs = b[bid].jcs;
	for( int i = 0; i < int( sf.size()); i++)
		sf[i].b[bid].rcs = b[bid].jcs;
}

// Description: Unmounts a given bone completely so it is no longer part of
//	the mount chain at all (whether it be a master or a slave).
//
//	Below is an example mount chain (A is the Master Mount Bone (mmb). B, C
//	and D are Mounted Slave Bones (msb)):
//
//		[A]---msb--->[B]---cms--->[C]---cms--->[D]
//               ^            |            |            |
//               +<----mmb----+            |            |
//               ^                         |            |
//               +<-----------mmb----------+            |
//               ^                                      |
//               +<---------------------mmb-------------+
//
//	All mounted slaves point back to the mmb through their respective
//	mmb pointers. Notice that the master accesses the first slave through
//	its msb pointer. From then on, it access all other slaves through
//	the previous slaves Chain Mounted Slave (cms) pointer. cms provides a
//	compact way for a master to reference all its slaves (inorder) without
//	having to allocate memory for slave pointers within its own structure.
//	cms also allows a bone to be a master mount bone and a mounted slave
//	bone at the same time. That is, the bone can access its own mounted
//	slaves through its msb, and the bone can allow its own master to access
//	its slaves through cms. DO NOT confuse cms with msb.
//
//	Here is an example showing mmb, msb and csm all in action (notice that
//	B is a mounted slave bone, and a master mount bone at the same time):
//
//                           +<--------------mmb-------+
//                           |+<---mmb----+            |
//                           ||           |            |
//                           ||+--msb--->[Q]---cms--->[R]
//                           |||
//		[A]---msb--->[B]---cms--->[C]---cms--->[D]
//               ^            |            |            |
//               +<----mmb----+            |            |
//               ^                         |            |
//               +<-----------mmb----------+            |
//               ^                                      |
//               +<---------------------mmb-------------+
//
//	Removing a bone from the mount chain is a lot like removing a node
//	from a double linked list.
//
void sk3DRomSkel::unmount_everything( const int bid, sk3DRomSkelFrame &p_sf)
{
	if( p_sf.b[bid].mmb >= 0) unmount_from_master( bid, p_sf);
	unmount_slaves( bid, p_sf);
}

// Description: Completely unmounts a given bone from the entire skeleton.
//	This is usually done when the bone is about to be deleted from the
//	skeleton.
void sk3DRomSkel::unmount_from_entire_skeleton( const int bid)
{
	unmount_everything( bid, skout);
	for( int i = 0; i < int( sf.size()); i++)
		unmount_everything( bid, sf[i]);
}

void sk3DRomSkel::get_bone_mount_slave_list( const int mmb,
	const sk3DRomSkelFrame &p_sf, std::vector <int> &blist)
{
	blist.clear();
	if( (mmb < 0) || (mmb >= int( b.size()))) return;
	int i = p_sf.b[mmb].msb;
	while( i >= 0) {
		blist.push_back( i);
		i = p_sf.b[i].cms;
	}
}

// Description: Unmounts all mounted slaves and destroys the cms mount chain
//	controlled by bone "bid" (all mounted slaves will become "free").
void sk3DRomSkel::unmount_slaves( const int bid, sk3DRomSkelFrame &p_sf)
{
	std::vector <int> mslist;
	get_bone_mount_slave_list( bid, p_sf, mslist);
	int i = mslist.size();
	while( (--i) >= 0) {
		unmount_bone( mslist[i], p_sf);
		mslist.erase( mslist.begin() + i);
	}
}

// Description: Unmounts a Mounted Slave Bone and removes it from the cms
//	mount chain (that is, bone "bid" will become "free").
void sk3DRomSkel::unmount_from_master( const int bid, sk3DRomSkelFrame &p_sf)
{
	const int mmb = p_sf.b[bid].mmb;
	p_sf.b[bid].mmb = -1;
#ifdef BONE_DEBUG
	if( (mmb < 0) || (mmb >= int( p_sf.b.size()))) {
		std::cerr << "Invalid mmb = " << mmb << ", Forcing SIGSEGV...";
		*((int *)0) = 0;
	}
#endif

	// If given bone is the first mounted slave.
	if( bid == p_sf.b[mmb].msb) {
		p_sf.b[mmb].msb = p_sf.b[bid].cms;
		return;
	}

	// Otherwise, the given bone must be a later mounted slave.
	int i = p_sf.b[mmb].msb;
	while( i >= 0) {
		if( bid == p_sf.b[i].cms) {
			p_sf.b[i].cms = p_sf.b[bid].cms;
			return;
		}
		i = p_sf.b[i].cms;
	}
}

void sk3DRomSkel::unmount_bone( const int bid, sk3DRomSkelFrame &p_sf)
{
	if( bone_id_is_invalid( bid)) return;
	if( p_sf.b[bid].mmb < 0) return;
	p_sf.b[bid].rcs = get_full_bone_rcs( bid, p_sf);
	unmount_from_master( bid, p_sf);
}

void sk3DRomSkel::unmount_bone_slaves( const int bid, sk3DRomSkelFrame &p_sf)
{
	if( (bid < 0) || (bid >= int( b.size()))) return;
	unmount_slaves( bid, p_sf);
}

// Description: Mounts a bone onto another bone. Only root bones may be
//	mounted. Naturally, a bone cannot be mounted onto itself. As well,
//	recursive mounts are not allowed. That is, if "B" is mounted onto "A",
//	then "A" cannot be mounted onto "B". In other words, if the caller
//	wishes to mount "A" onto "B", then "A" must NOT be a part of "B"'s
//	mount chain (that is, if "B" is mounted onto "C" and "C" is mounted
//	onto "A", then "A" cannot be mounted onto "B" or "C").
void sk3DRomSkel::mount_bone( const int tid, const int bid,
	sk3DRomSkelFrame &p_sf)
{
	if( bone_id_is_invalid( bid)) return;
	if( (tid < 0) || (tid >= int( b.size()))) return;
	if( tid == bid) return;
	if( b[bid].mid >= 0) return;
	if( bone_is_in_mount_chain( bid, tid, p_sf)) return;

	unmount_bone( bid, p_sf);
	
	rgCoordSystem mcs( get_full_bone_rcs( tid, p_sf));
	p_sf.b[bid].rcs = get_full_bone_rcs( bid, p_sf);
	bind_rcs_to_cs( p_sf.b[bid].rcs, mcs);

	p_sf.b[bid].mmb = tid;
	if( p_sf.b[tid].msb < 0) p_sf.b[tid].msb = bid;
	else {
		int i = p_sf.b[tid].msb;
		while( p_sf.b[i].cms >= 0) i = p_sf.b[i].cms;
		p_sf.b[i].cms = bid;
	}
}

bool sk3DRomSkel::bone_is_in_mount_chain( const int bid, const int cid,
	sk3DRomSkelFrame &p_sf)
{
	if( bid < 0) return false;
	if( cid < 0) return false;
	// If bone cid has no Master Mount Bone (mmb), then no bone is part of
	// its mount chain.
	if( p_sf.b[cid].mmb < 0) return false;
	int i = p_sf.b[cid].mmb;
	if( bid == i) return true;
	i = p_sf.b[i].msb;
	while( i >= 0) {
		if( bid == i) return true;
		i = p_sf.b[i].cms;
	}
	return false;
}

void sk3DRomSkel::unbind_bone( const int bid)
{
	if( (bid < 0) || (bid >= int( b.size()))) return;
	sk3DRomBone &bn = b[bid];
	if(  bn.mid < 0) return;
	unbind_bone_jcs( bid);
	sk3DRomBone &mb = b[ bn.mid];
	const int istop = int( mb.link.size());
	for( int i = 0; i < istop; i++) {
		if( bid == mb.link[i].bid) {
			mb.link.erase( mb.link.begin() + i);
			break;
		}
	}
	bn.mid = -1;
	add_root_bone( bid);
	build_master_list();
}

void sk3DRomSkel::unbind_bone_slaves( sk3DRomBone &p_bn)
{
	while( !(p_bn.link.empty())) unbind_bone( p_bn.link[0].bid);
}

// Description: Binds a bone to it's Joint Coordinate System (jcs). If a
//	valid Master Coordinate System (mcs) is also given, then the bone's
//	jcs will be bound to the mcs. All skeleton frames (including  "skout"
//	will have thier rcs's updated to reflect the binding).
//
// Parameters: "mcs" is the Master Coordinate System, and "bn" is the bone to
//	bind.
void sk3DRomSkel::bind_bone_to_jcs( rgCoordSystem *mcs, sk3DRomBone &bn)
{
	bn.ja = bn.jcs.change_point_csys( bn.a);
	bn.jb = bn.jcs.change_point_csys( bn.b);
	if( mcs) bn.jcs = mcs->change_csys_csys( bn.jcs);
}

// Description: Unbinds a bone's Joint Coordinate System (jcs).
void sk3DRomSkel::unbind_bone_jcs( const int bid)
{
	b[bid].jcs = get_full_bone_jcs( bid);
	skout.b[bid].rcs = get_full_bone_rcs( bid, skout);
	const int istop = int( sf.size());
	for( int i = 0; i < istop; i++)
		sf[i].b[bid].rcs = get_full_bone_rcs( bid, sf[i]);
}

void sk3DRomSkel::init_skel_frame( sk3DRomSkelFrame &p_sf)
{
	p_sf.b.clear(); // Delete all current bones (if any).
	// Create new bones in the default orientation.
	const int istop = int( b.size());
	for( int i = 0; i < istop; i++) {
		p_sf.b.push_back( sk3DRomBoneFrame());
		p_sf.b[i].rcs = b[i].jcs;
	}
}

// Description: Add a new bone to the skeleton without disturbing the current
//	skeleton or any existing skeleton frames. The new bone added is also
//	a root bone.
//
// What is done:
//	1. Bone is added to main skeleton.
//	2. Skeleton output is updated to include new bone.
//	3. All existing skeleton frames are updated to include new bone.
sk3DRomBone *sk3DRomModel::add_bone( rgVector ja, rgVector jb)
{
	if( (ja - jb).length() < 1.0f) return 0;
	skel.b.push_back( sk3DRomBone());
	sk3DRomBone *b = &(skel.b[ skel.b.size() - 1]);
	skel.position_bone( *b, ja, jb);
	skel.rb.push_back( skel.b.size() - 1);

	sk3DRomBoneFrame bf;
	bf.rcs = b->jcs;
	skel.skout.b.push_back( bf);

	const int istop = int( skel.sf.size());
	for( int i = 0; i < istop; i++) skel.sf[i].b.push_back( bf);

	return b;
}

void sk3DRomSkel::position_bone( sk3DRomBone &p_b, const rgVector ja,
	const rgVector jb)
{
	p_b.a = ja;
	p_b.b = jb;
	create_bone_jcs( p_b);
	bind_bone_to_jcs( 0, p_b);
}

// TODO: Currently "reposition_bone" does not support repositioning of bones
//	that have slaves or mounted slaves. If time permits, this feature
//	should be added as it will provide convinience to the modeler.
//	However, the current simple solution around this is to force the
//	modeler to only reposition bones that have no slaves or mounted
//	slaves. This is reasonable since the modeler shouldn't be repositioning
//	bones of a bound and mounted skeleton anyway (since it is not trivial
//	to know how the repositioning will affect the skeleton frames).
//	XXX: As well, it does not currently support bones with masters.
void sk3DRomModel::reposition_bone( const int bid, const rgVector ja,
	const rgVector jb)
{
	if( (bid < 0) || (bid >= int( skel.b.size()))) return;
	unmap_bone_vertices( bid, 0);
	skel.position_bone( skel.b[bid], ja, jb);
	bind_bone_verticies( bid);
}

void sk3DRomSkel::get_bone_slave_list( const int bid, std::vector <int> &blist)
{
	blist.clear();
	const int istop = int( b[bid].link.size());
	for( int i = 0; i < istop; i++)
		blist.push_back( b[bid].link[i].bid);
}

void sk3DRomModel::notify_verts_of_bone_swap( const int old_bid,
	const int new_bid)
{
	const std::vector <short> &vid = skel.b[ old_bid].vid;
	const int istop = int( skel.b[ old_bid].vid.size());
	for( int i = 0; i < istop; i++) v[ vid[i]].m_bid = new_bid;
}

// Description:
//
//	XXX This function MUST ONLY BE CALLED from "delete_bone".
//	XXX "new_bid" MUST be a bone that was marked for deletion. "old_bid"
//	    will take over the index of "new_bid", so the bone that use to
//	    own "new_bid" MUST be deleted.
void sk3DRomModel::notify_skel_of_bone_id_change( const int old_bid,
	const int new_bid)
{
	skel.rb.clear();
	notify_verts_of_bone_swap( old_bid, new_bid);
	sk3DRomBone &bold = skel.b[ old_bid];
	sk3DRomBone &bnew = skel.b[ new_bid];

	// Notify master.
	if( bold.mid >= 0) {
		sk3DRomBone &m = skel.b[ bold.mid];
		const int istop = m.link.size();
		int i = 0;
		for( ; i < istop; i++) {
			if( old_bid == m.link[i].bid) {
				m.link[i].bid = new_bid;
				break;
			}
		}
		for( ++i; i < istop; i++) {
			if( new_bid == m.link[i].bid) {
				m.link.erase( m.link.begin() + i);
				break;
			}
		}
	}

	// Notify the slaves.
	{
		const int istop = int( bold.link.size());
		for( int i = 0; i < istop; i++)
			skel.b[ bold.link[i].bid].mid = new_bid;
	}

	// Notify all skeleton frames.
	{
		skel.skout.b[ new_bid] = skel.skout.b[ old_bid];
		const int istop = int( skel.sf.size());
		for( int i = 0; i < istop; i++)
			skel.sf[i].b[ new_bid] = skel.sf[i].b[ old_bid];
	}

	// XXX: "new_bid" WILL NOT be referenced ANYWHERE within the mount
	//	chain of each skeleton. This is because the bone that used to
	//	own "new_bid" is being deleted (and thus would have been
	//	unmounted previously).
	skel.notify_mounts_of_bone_id_change( old_bid, new_bid);

	bnew = bold;

	skel.build_master_list();
}

void sk3DRomSkel::change_skel_mount_bone_id( const int old_bid,
	const int new_bid, sk3DRomSkelFrame &p_sf)
{
	for( int i = 0; i < int( p_sf.b.size()); i++) {
		if( old_bid == p_sf.b[i].mmb) p_sf.b[i].mmb = new_bid;
		if( old_bid == p_sf.b[i].msb) p_sf.b[i].msb = new_bid;
		if( old_bid == p_sf.b[i].cms) p_sf.b[i].cms = new_bid;
	}
}

// Description: Updates the mount chain for the entire skeleton when a bone
//	is having its index number changed.
void sk3DRomSkel::notify_mounts_of_bone_id_change( const int old_bid,
	const int new_bid)
{
	change_skel_mount_bone_id( old_bid, new_bid, skout);
	for( int i = 0; i < int( sf.size()); i++)
		change_skel_mount_bone_id( old_bid, new_bid, sf[i]);
}

// Description: Deletes a bone from the skeleton. Slaves of the bone to delete
//	will become root bones.
//
// Vertices: Vertices will be unbound using the coordinate systems of the
//	skeleton frame "sf". If no skeleton frame is provided then the vertices
//	will be unbound using the default jcs coordinate system.
//
// Bone:
//	- mid (Index to master)
//	- mlist (List of master indicies)
//	- link (List of slave bones)
//	- vid (List of controlled vertcies)
//
// Vertices:
//	- m_bid (Indexes bones which control it)
//
//
//
void sk3DRomModel::delete_bone( const int bid)
{
	delete_bone( bid, 0);
}

void sk3DRomModel::delete_bone( const int bid, sk3DRomSkelFrame *sf)
{
	if( (bid < 0) || (bid >= int( skel.b.size()))) return;

	skel.unmount_from_entire_skeleton( bid);
	unbind_bone_vertices( bid, sf);
	skel.unbind_bone_slaves( skel.b[bid]);
	skel.unbind_bone( bid);

	const int last = int( skel.b.size()) - 1;
	if( bid != last) notify_skel_of_bone_id_change( last, bid);
	skel.b.erase( skel.b.begin() + last);
	skel.skout.b.erase( skel.skout.b.begin() + last);
	const int istop = int( skel.sf.size());
	for( int i = 0; i < istop; i++)
		skel.sf[i].b.erase( skel.sf[i].b.begin() + last);
	skel.build_root_bone_list();
	skel.build_master_list();
}

void sk3DRomModel::bind_bone_verticies( const short bid)
{
	sk3DRomBone &b = skel.b[bid];
	rgCoordSystem jcs = skel.get_full_bone_jcs( bid);
	const short istop = short( b.vid.size());
	for( short i = 0; i < istop; i++) {
		v[ b.vid[i]] = jcs.change_point_csys( v[ b.vid[i]]);
		v[ b.vid[i]].m_bid = bid;
	}
}

void sk3DRomModel::bind_vertex_to_bone( const int vid, const int bid)
{
	bind_vertices_to_bone( &vid, 1, bid);
}

void sk3DRomModel::bind_vertices_to_bone( const int *vlist, int count,
        const int bid)
{

	// NOTE: The vertices in "v" referenced/indexed by "vlist" are
	//      expected to be relative to the standard coordinate system
	//      (X = (1, 0, 0), Y = (0, 1, 0), Z = (0, 0, 1), origin =
	//      (0, 0, 0)).
	if( !vlist) return;
	if( count <= 0) return;
	if( (bid < 0) || (bid >= int( skel.b.size()))) return;
	unbind_vertices( vlist, count);
	sk3DRomBone &b = skel.b[ bid];
	rgCoordSystem rcs = skel.get_full_bone_rcs( bid, skel.skout);
	while( count--) {
		v[ *vlist] = rcs.change_point_csys( v[ *vlist]);
		v[ *vlist].m_bid = bid;
		b.vid.push_back( *vlist);
		vlist++;
	}
	build_bone_less_list();
}

void sk3DRomModel::unbind_vertex( const int vid)
{
	if( (vid < 0) || (vid >= int( v.size()))) return;
	if( v[ vid].bid() < 0) return;
	rgCoordSystem jcs( skel.get_full_bone_jcs( v[ vid].bid()));
	unbind_vertex( vid, &jcs);
	build_bone_less_list();
}

void sk3DRomModel::unbind_vertex( const int vid, rgCoordSystem *rcs)
{
	sk3DRomBone &b = skel.b[ v[ vid].bid()];
	const int istop = int( b.vid.size());
	for( int i = 0; i < istop; i++) {
		if( vid == b.vid[i]) {
			b.vid.erase( b.vid.begin() + i);
			break;
		}
	}
	v[ vid] = rcs->map_point( v[ vid]);
	v[ vid].m_bid = -1;
}

void sk3DRomModel::unbind_vertices( const int *vlist, int count)
{
	if( !vlist) return;
	if( count <= 0) return;
	if( skel.b.empty()) return;

	// Prepare full rcs's for all bones referenced.
	rgCoordSystem *rcs = new rgCoordSystem[ skel.b.size()];
	char *ready = new char[ skel.b.size()];
	memset( ready, 0, skel.b.size());
	for( int i = 0; i < count; i++) {
		{
			const int bid = v[ vlist[i]].bid();
			if( (bid >= 0) && !( ready[bid])) {
				rcs[bid] = skel.get_full_bone_rcs( bid,
					skel.skout);
				ready[bid] = 1;
			}
		}
	}
	delete [] ready;

	while( count--) {
		if( v[ *vlist].bid() >= 0)
			unbind_vertex( *vlist, rcs + v[ *vlist].bid());
		vlist++;
	}
	delete [] rcs;
	build_bone_less_list();
}

void sk3DRomModel::unmap_bone_vertices( const int bid, sk3DRomSkelFrame *sf)
{
	if( (bid < 0) || (bid >= int( skel.b.size()))) return;
	sk3DRomBone &b = skel.b[bid];
	rgCoordSystem cs;
	if( sf) cs = skel.get_full_bone_rcs( bid, *sf);
	else cs = skel.get_full_bone_jcs( bid);
	const int istop = int( b.vid.size());
	for( int i = 0; i < istop; i++) {
		v[ b.vid[i]] = cs.map_point( v[ b.vid[i]]);
		v[ b.vid[i]].m_bid = -1;
	}
}

void sk3DRomModel::unbind_bone_vertices( const int bid, sk3DRomSkelFrame *sf)
{
	unmap_bone_vertices( bid, sf);
	skel.b[bid].vid.clear();
}

/*
void sk3DRomSkel::break_bone_master_slave_link( const int bid)
{
	if( bid < 0) return;

	// Remove slave's link to the master.
	const int mid = b[bid].mid;
	if( mid < 0) return;
	b[bid].mid = -1;
	create_bone_jcs( b[bid]);
	bind_bone_to_jcs( 0, b[bid]);

	// Remove master's link to the slave.
	sk3DRomBone &m = b[mid];
	const int istop = int( m.link.size());
	for( int i = 0; i < istop; i++) {
		if( bid == m.link[i].bid) {
			m.link.erase( m.link.begin() + i);
			return;
		}
	}
}
*/

// FIXME: The absolute vertex generation code is duplicated in rg3DRom_render
//	by the "render_skel_frame" supporting/helper functions. They should
//	be removed and the absolute vertex generation code supplied here
//	should be used instead.
inline void hlp_gen_absolute_bone_vertices( const sk3DRomSkel &s,
	const sk3DRomSkelFrame &sf, rgCoordSystem *mcs,
	const int bid, rgVector &ja, rgVector &jb)
{
#ifdef BONE_DEBUG
	if( (bid < 0) || (bid >= int( sf.b.size()))) {
		std::cerr << "Invalid bid = " << bid <<
			"!!! Forcing SIGSEGV...";
		*((int *)0) = 0;
	}
#endif
	rgCoordSystem cs( sf.b[bid].rcs);
	if( mcs) cs = mcs->map_csys( cs);
	const sk3DRomBone &b = s.b[bid];
	ja = cs.map_point( b.ja);
	jb = cs.map_point( b.jb);
}

struct MiniBone
{ 
	rgVector a;
	rgVector b;
};

void hlp_gen_absolute_slave_bone_vertices( const sk3DRomSkel &s,
	const sk3DRomSkelFrame &sf, rgCoordSystem *mcs, const int bid,
	MiniBone *mb)
{
	rgCoordSystem cs( sf.b[bid].rcs);
	if( mcs) cs = mcs->map_csys( cs);
	const std::vector <sk3DRomBoneLink> &link = s.b[bid].link;
	const int istop = int( link.size());
	for( int i = 0; i < istop; i++) {

		hlp_gen_absolute_bone_vertices( s, sf, &cs, link[i].bid,
			mb[ link[i].bid].a, mb[ link[i].bid].b);

		if( !(s.b[ link[i].bid].link.empty())) {
			hlp_gen_absolute_slave_bone_vertices( s, sf, &cs,
				link[i].bid, mb);
		}
	}
}

// TODO: Optimize this (though not neccasary) to use ordered root bones
//	with mounts.
//
void hlp_gen_absolute_skel_vertices( sk3DRomSkel &s,
	const sk3DRomSkelFrame &sf, MiniBone *mb)
{
	/*
	for( int i = 0; i < int( s.rb.size()); i++) {
		hlp_gen_absolute_bone_vertices( s, sf, 0, s.rb[i],
			mb[ s.rb[i]].a, mb[ s.rb[i]].b);
		hlp_gen_absolute_slave_bone_vertices( s, sf, 0, s.rb[i], mb);
	}
	*/
	const int istop = int( s.b.size());
	for( int i = 0; i < istop; i++) {
		{
			rgCoordSystem cs( s.get_full_bone_rcs( i, sf));
			mb[i].a = cs.map_point( s.b[i].ja);
			mb[i].b = cs.map_point( s.b[i].jb);
		}
	}
}

// Description: Performs an orthographic selection of a bone from a given
//	skeleton frame. This is done by taking the frame and transforming
//	all of its bones to produce the bones absolute vertex locations.
//	Selection is then done against the absolute vertex locations.
void sk3DRomModel::ortho_select_bone( rgVector p, const rgVector n,
        const sk3DRomSkelFrame &sf, int &bid, int &jid)
{
	ortho_select_bone( p, n, sf, bid, jid, 32.0f);
}

void sk3DRomModel::ortho_select_bone( rgVector p, const rgVector n,
	const sk3DRomSkelFrame &sf, int &bid, int &jid, const float min)
{
	bid = -1;
	jid = -1;

	MiniBone *mb = new MiniBone[ sf.b.size()];
	hlp_gen_absolute_skel_vertices( skel, sf, mb);

	p = rgMakePlanar( p, n);

	float d = -1.0f;
	const int istop = int( skel.b.size());
	for( int i = 0; i < istop; i++) {

		// Flatten bone vertices since an orthographic search is
		// being done.
		const rgVector u( rgMakePlanar( mb[i].a, n));
		const rgVector w( rgMakePlanar( mb[i].b, n));

		// If the selection point "p" is between the two bone joints
		const rgVector bone_normal( (n.cross( u - w)).normalize());
		if(	(((u - p).dot( w - p)) < 0.0f) &&
			(fabs( bone_normal.dot( u - p)) <= min)) {
			const float t = (u - p).length();
			const float k = (w - p).length();
			// If joint "a" is closer than "b", or equal to
			if( t <= k) {
				// If closest joint so far
				if( (t < d) || (d < 0.0f)) {
					bid = i;
					jid = 0;
					d = t;
				}

			// Else, joint "b" is closer than "a"
			} else {
			// If closest joint so far
				if( (k < d) || (d < 0.0f)) {
					bid = i;
					jid = 1;
					d = k;
				}
			}
		}
	}

	delete [] mb;
}

int sk3DRomModel::get_absolute_bone_vertices( const sk3DRomSkelFrame &sf,
	const int bid, rgVector &a, rgVector &b)
{
	if( (bid < 0) || (bid >= int( sf.b.size()))) return -1;
	rgCoordSystem mcs;
	skel.get_full_bone_rcs( bid, sf, mcs);
	hlp_gen_absolute_bone_vertices( skel, sf, &mcs, bid, a, b);
	return skel.b[bid].pj;
}

void sk3DRomModel::add_skeleton_frame( const sk3DRomSkelFrame &s)
{
	skel.sf.push_back( s);
}



