/*
Copyright (C) 2003-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "rgChaikin.h"

void hlp_chaikin_sub_divide( std::vector <rgVector> &f,
	std::vector <rgVector> &c);

int rg_chaikin_sub_divide( std::vector <rgVector> &f,
	std::vector <rgVector> &c, int level)
{
	if( &f == &c) return -1;
	if( c.empty()) {
		f.clear();
		return 0;
	}
	if( level <= 0) {
		f = c;
		c.clear();
		return 0;
	}

	f.clear();
	std::vector <rgVector> *src = &c;
	std::vector <rgVector> *des = &f;
	std::vector <rgVector> *tmp;
	while( level--) {
		hlp_chaikin_sub_divide( *des, *src);
		tmp = src;
		src = des;
		des = tmp;
	}

	if( f.empty()) {
		f = c;
		c.clear();
	}

	return 0;
}

void hlp_chaikin_sub_divide( std::vector <rgVector> &f,
	std::vector <rgVector> &c)
{
	const int i_stop = c.size() - 1;
	f.push_back( c[0]);		// First point
	for( int i = 1; i < i_stop; i++) {
		f.push_back( ((c[i - 1] * 0.25f) + (c[i] * 0.75f)));
		f.push_back( ((c[i] * 0.75f) + (c[i + 1] * 0.25f)));
	}
	f.push_back( c[ c.size() - 1]);	// Last point
	c.clear();
}
