/*
Copyright (C) 1999-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "rgStandard.h"

rgVector::rgVector()
{
	x = 0.0f;
	y = 0.0f;
	z = 0.0f;
}

rgVector::rgVector( rgFloat p_x, rgFloat p_y, rgFloat p_z)
{
	x = p_x;
	y = p_y;
	z = p_z;
}

std::ostream &operator <<( std::ostream &o, rgVector v)
{
	return o << "( " << v.x << ", " << v.y << ", " << v.z << ")";
}

rgVector &rgVector::normalize()
{
	rgFloat len = length();
	// XXX: Only vectors with non-zero length are normalized
	if( 0.0f == len) return *this;
	
	len = 1.0f/len;
	x *= len;
	y *= len;
	z *= len;

	return *this;
}

//bool rgVector::null() const
//{
//	if( (0.0f == x) && (0.0f == y) && (0.0f == z)) return true;
//	return false;
//}


/***************************************************************************
 * rgBasis
 ***************************************************************************/
rgBasis::rgBasis()
{
	side =	rgVector( 1, 0, 0);	/* "X" axis */
	up =	rgVector( 0, 1, 0);	/* "Y" axis */
	eye =	rgVector( 0, 0, 1);	/* "Z" axis */
}

std::ostream &operator <<( std::ostream &o, rgBasis &b)
{
	return o << "{ side = " << b.side << "\n\tup = " << b.up
		<< "\n\teye = " << b.eye << "}";
}

/*
rgVector rgCoordSystem::change_point_csys( rgVector point, rgCoordSystem &old)
{
	return change_vector_csys( point, old) + p;
}

rgVector rgCoordSystem::change_vector_csys( rgVector v, rgCoordSystem &old)
{
	v = v - old.p;
	return	(v.dot( old.side) * side) +
		(v.dot( old.up) * up) +
		(v.dot( old.eye) * eye);
}
*/

/*
rgCoordSystem rgCoordSystem::map_csys( rgCoordSystem p_cs)
{
	rgCoordSystem cs;
	cs.p = map_point( p_cs.p);
	cs.side = map_vector( p_cs.side);
	cs.up = map_vector( p_cs.up);
	cs.eye = map_vector( p_cs.eye);
	return cs;
}
*/

void rgCoordSystem::map_coord_system( rgCoordSystem &dcs, rgCoordSystem &scs)
{
	dcs.p = map_point( scs.p);
	dcs.side = map_vector( scs.side);
	dcs.up = map_vector( scs.up);
	dcs.eye = map_vector( scs.eye);
}

//rgVector rgCoordSystem::change_point_csys( rgVector q)
//{
//	return change_vector_csys( q - p);
//}


//rgVector rgCoordSystem::change_vector_csys( const rgVector v)
//{
//	return	rgVector( side.dot( v), up.dot( v), eye.dot( v));
//}


rgCoordSystem rgCoordSystem::change_csys_csys( const rgCoordSystem p_cs) const
{
	rgCoordSystem cs;
	cs.p = change_point_csys( p_cs.p);
	cs.side = change_vector_csys( p_cs.side);
	cs.up = change_vector_csys( p_cs.up);
	cs.eye = change_vector_csys( p_cs.eye);
	return cs;
}

std::ostream &operator <<( std::ostream &o, const rgCoordSystem &c)
{
	return o	<< "{\torigin = " << c.p
			<< "\n\tside = " << c.side
			<< "\n\tup = " << c.up
			<< "\n\teye = " << c.eye
			<< "}\n";
}

rgMatrix::rgMatrix()
{
    loadIdentity();
}

rgMatrix::rgMatrix( const rgCoordSystem &cs)
{
	convert_csys_to_matrix( cs);
}

void rgMatrix::convert_csys_to_matrix( const rgCoordSystem &cs)
{
	d[0] = cs.side.x;
	d[1] = cs.side.y;
	d[2] = cs.side.z;
	d[3] = 0;

	d[4] = cs.up.x;
	d[5] = cs.up.y;
	d[6] = cs.up.z;
	d[7] = 0;

	d[8] = cs.eye.x;
	d[9] = cs.eye.y;
	d[10] = cs.eye.z;
	d[11] = 0;

	d[12] = 0;
	d[13] = 0;
	d[14] = 0;
	d[15] = 1.0f;

	rgMatrix t;
	t.translate( cs.p);
	(*this) = (*this) * t;
}

rgMatrix &rgMatrix::apply_positioning_csys( const rgCoordSystem &cs)
{
        rgMatrix t;
	t.translate( cs.p);
	rgMatrix m;
	// FIXME: Why does the current code work and the commented out version
	//	not work???
	//m.d[0] = cs.side.x; m.d[1] = cs.side.y; m.d[2] = cs.side.z;
	//m.d[4] = cs.up.x; m.d[5] = cs.up.y; m.d[6] = cs.up.z;
	//m.d[8] = cs.eye.x; m.d[9] = cs.eye.y; m.d[10] = cs.eye.z;
	m.d[0] = cs.side.x; m.d[1] = cs.up.x; m.d[2] = cs.eye.x;
	m.d[4] = cs.side.y; m.d[5] = cs.up.y; m.d[6] = cs.eye.y;
	m.d[8] = cs.side.z; m.d[9] = cs.up.z; m.d[10] = cs.eye.z;
	return (*this = t * m);
}

void rgMatrix::loadIdentity()
{
    /* Set the matrix to the indentity matrix */
    d[0] = 1.0f; d[1] = 0.0f; d[2] = 0.0f; d[3] = 0.0f;
    d[4] = 0.0f; d[5] = 1.0f; d[6] = 0.0f; d[7] = 0.0f;
    d[8] = 0.0f; d[9] = 0.0f; d[10] = 1.0f; d[11] = 0.0f;
    d[12] = 0.0f; d[13] = 0.0f; d[14] = 0.0f; d[15] = 1.0f;    
}

rgMatrix rgMatrix::operator +( const rgMatrix &m) const
{
    rgMatrix n;
    n = *this;
    n += m;
    return n;
}

rgMatrix rgMatrix::operator -( const rgMatrix &m) const
{
    rgMatrix n;
    n = *this;
    n -= m;
    return n;
}

rgMatrix &rgMatrix::operator +=( const rgMatrix &m)
{
    d[0] += m.d[0]; d[4] += m.d[4];
    d[1] += m.d[1]; d[5] += m.d[5];
    d[2] += m.d[2]; d[6] += m.d[6];
    d[3] += m.d[3]; d[7] += m.d[7];
    
    d[8] += m.d[8]; d[12] += m.d[12];
    d[9] += m.d[9]; d[13] += m.d[13];
    d[10] += m.d[10]; d[14] += m.d[14];
    d[11] += m.d[11]; d[15] += m.d[15];
    
    return *this;
}

rgMatrix &rgMatrix::operator -=( const rgMatrix &m)
{
    d[0] -= m.d[0]; d[4] -= m.d[4];
    d[1] -= m.d[1]; d[5] -= m.d[5];
    d[2] -= m.d[2]; d[6] -= m.d[6];
    d[3] -= m.d[3]; d[7] -= m.d[7];
    
    d[8] -= m.d[8]; d[12] -= m.d[12];
    d[9] -= m.d[9]; d[13] -= m.d[13];
    d[10] -= m.d[10]; d[14] -= m.d[14];
    d[11] -= m.d[11]; d[15] -= m.d[15];
    
    return *this;
}

rgMatrix &rgMatrix::operator *=( const rgFloat k)
{
    d[0] *= k; d[1] *= k; d[2] *= k; d[3] *= k;
    d[4] *= k; d[5] *= k; d[6] *= k; d[7] *= k;
    d[8] *= k; d[9] *= k; d[10] *= k; d[11] *= k;
    d[12] *= k; d[13] *= k; d[14] *= k; d[15] *= k;
    
    return *this;
}

rgMatrix &rgMatrix::operator *=( const rgMatrix &m)
{   
    /* Must use a new matrix "n" to store the results since each element of
      * "d" must be preserved until the entire matrix has been multiplied.
      */
    rgMatrix n;
    n.d[0] = d[0]*m.d[0] + d[1]*m.d[4] + d[2]*m.d[8] + d[3]*m.d[12];
    n.d[1] = d[0]*m.d[1] + d[1]*m.d[5] + d[2]*m.d[9] + d[3]*m.d[13];
    n.d[2] = d[0]*m.d[2] + d[1]*m.d[6] + d[2]*m.d[10] + d[3]*m.d[14];
    n.d[3] = d[0]*m.d[3] + d[1]*m.d[7] + d[2]*m.d[11] + d[3]*m.d[15];
    
    n.d[4] = d[4]*m.d[0] + d[5]*m.d[4] + d[6]*m.d[8] + d[7]*m.d[12];
    n.d[5] = d[4]*m.d[1] + d[5]*m.d[5] + d[6]*m.d[9] + d[7]*m.d[13];
    n.d[6] = d[4]*m.d[2] + d[5]*m.d[6] + d[6]*m.d[10] + d[7]*m.d[14];
    n.d[7] = d[4]*m.d[3] + d[5]*m.d[7] + d[6]*m.d[11] + d[7]*m.d[15];
    
    n.d[8] = d[8]*m.d[0] + d[9]*m.d[4] + d[10]*m.d[8] + d[11]*m.d[12];
    n.d[9] = d[8]*m.d[1] + d[9]*m.d[5] + d[10]*m.d[9] + d[11]*m.d[13];
    n.d[10] = d[8]*m.d[2] + d[9]*m.d[6] + d[10]*m.d[10] + d[11]*m.d[14];
    n.d[11] = d[8]*m.d[3] + d[9]*m.d[7] + d[10]*m.d[11] + d[11]*m.d[15];
    
    n.d[12] = d[12]*m.d[0] + d[13]*m.d[4] + d[14]*m.d[8] + d[15]*m.d[12];
    n.d[13] = d[12]*m.d[1] + d[13]*m.d[5] + d[14]*m.d[9] + d[15]*m.d[13];
    n.d[14] = d[12]*m.d[2] + d[13]*m.d[6] + d[14]*m.d[10] + d[15]*m.d[14];
    n.d[15] = d[12]*m.d[3] + d[13]*m.d[7] + d[14]*m.d[11] + d[15]*m.d[15];
    
    *this = n;
    return *this;
}

std::ostream &operator <<( std::ostream &o, const rgMatrix &m)
{
	return
		o << "| " << m.d[0] << ", " << m.d[1] << ", " << m.d[2] << ", "
		<< m.d[3] << " |\n"
		<< "| " << m.d[4] << ", " << m.d[5] << ", " << m.d[6] << ", "
		<< m.d[7] << " |\n"
		<< "| " << m.d[8] << ", " << m.d[9] << ", " << m.d[10]
		<< ", " << m.d[11] << " |\n"
		<< "| " << m.d[12] << ", " << m.d[13] << ", " << m.d[14]
		<< ", " << m.d[15] << " |";
}

void rgMatrix::rotate( const rgFloat ang, const rgFloat x, const rgFloat y,
	const rgFloat z)
{
    /* Pre-calculating results that will be used repeatedly */
    const rgFloat cAng = (float)cos(ang);
    const rgFloat oneMinusCAng = 1.0f - cAng;
    const rgFloat sinAng = (float)sin(ang);

    /* Calculate the rotation matrix */
    rgMatrix n;	/* Create a new identity matrix */
    n.d[0] = x * x * oneMinusCAng + cAng;
    n.d[1] = x * y * oneMinusCAng - z * sinAng;
    n.d[2] = x * z * oneMinusCAng + y * sinAng;
    
    n.d[4] = x * y * oneMinusCAng + z * sinAng;
    n.d[5] = y * y * oneMinusCAng + cAng;
    n.d[6] = y * z * oneMinusCAng - x * sinAng;

    n.d[8] = x * z * oneMinusCAng - y * sinAng;
    n.d[9] = y * z * oneMinusCAng + x * sinAng;
    n.d[10] = z * z * oneMinusCAng + cAng;
    
    /* Apply the rotation to "this" matrix */
    *(this) = (*this) * n;
}

void rgMatrix::scale( rgVector v)
{
	scale( v.x, v.y, v.z);
}

void rgMatrix::scale( const rgFloat x, const rgFloat y, const rgFloat z)
{
	rgMatrix n;
	n.d[0] = x;
	n.d[5] = y;
	n.d[10] = z;

	*(this) = (*this) * n;
}

void rgMatrix::translate( rgVector v)
{
	translate( v.x, v.y, v.z);
}

void rgMatrix::translate( const rgFloat x, const rgFloat y, const rgFloat z)
{
    /* Create a new identity matrix */
    rgMatrix n;
    
    n.d[3] = x;
    n.d[7] = y;
    n.d[11] = z;
    
    /* Apply the translation to "this" matrix */
    *(this) = (*this) * n;
}

bool rgMatrix::frustum( rgFloat left, rgFloat right, rgFloat bottom,
	rgFloat top, rgFloat zNear, rgFloat zFar)
{
	const rgFloat rightMinusLeft = right - left;
	const rgFloat topMinusBottom = top - bottom;
    
	if( 0.0f == rightMinusLeft) return false;
	if( 0.0f == topMinusBottom) return false;
	if( (zNear < 0.0f) || (zFar < 0.0f)) return false;
    
	const rgFloat zFarMinusZnear = zFar - zNear;
	if( 0.0f == zFarMinusZnear) return false;
    
	rgMatrix m;
	m.d[0] = 2.0f*zNear/rightMinusLeft;
	m.d[2] = (right + left)/rightMinusLeft;
	m.d[6] = (top + bottom)/topMinusBottom;
	m.d[5] = 2.0f*zNear/topMinusBottom;
	m.d[10] = -(zFar + zNear)/zFarMinusZnear;
	m.d[11] = -(2.0f*zFar*zNear)/zFarMinusZnear;
	m.d[14] = -1.0f;
	m.d[15] = 0.0f;   
 
	*(this) *= m;
    
	return true;
}

bool rgMatrix::ortho( rgFloat left, rgFloat right, rgFloat bottom, rgFloat top,
	rgFloat zNear, rgFloat zFar)
{
	const rgFloat rightMinusLeft = right - left;
	const rgFloat topMinusBottom = top - bottom;
	const rgFloat zFarMinusZnear = zFar - zNear;

	if( 0.0f == rightMinusLeft) return false;
	if( 0.0f == topMinusBottom) return false;
	if( 0.0f == zFarMinusZnear) return false;

	rgMatrix m;
	m.d[0] = 2.0f/rightMinusLeft;
	m.d[3] = -(right + left)/rightMinusLeft;
	m.d[5] = 2.0f/topMinusBottom;
	m.d[7] = -(top + bottom)/topMinusBottom;
	m.d[10] = -2.0f/zFarMinusZnear;
	m.d[11] = -(zFar + zNear)/zFarMinusZnear;
	m.d[15] = 1.0f;
	*(this) *= m;
    
	return true;
}

bool rgMatrix::perspective( rgFloat fovy, rgFloat aspect, rgFloat zNear,
	rgFloat zFar)
{
	return perspective( fovy, aspect, zNear, zFar, 0, 0);
}

bool rgMatrix::perspective( rgFloat fovy, rgFloat aspect, rgFloat zNear,
	rgFloat zFar, rgFloat *p_width, rgFloat *p_height)
{
	if( zNear <= 0.0f) return false;
	if( zFar <= 0.0f) return false;

	/* Convert the degrees into radians */
	fovy = RG_DEGTORAD( fovy);

	/* Maximum supported field of view is < 180.0f degrees */
	const rgFloat theta = 0.5f * fovy;
	if( theta >= RG_90_DEG_IN_RAD) return false;

	/* Determine the scaling required to produce a near plane that will
	 * give the requested field of view.
	 */
	rgFloat yScale = zNear * tan( theta);
	if( 0.0f == yScale) return false;
	yScale = 1.0f/yScale;

	/* Create the requested projection matrix */
	rgMatrix p;
	p.scale( rgVector( yScale * aspect, yScale, 1.0f));
	if( !p.frustum( -1.0f, 1.0f, -1.0f, 1.0f, zNear, zFar))
		return false;

	if( p_width) *p_width = yScale * aspect;
	if( p_height) *p_height = yScale;

	/* Apply the projection */
	*(this) *= p;
	return true;
}



#define MOV_MATRIX_ROW( des, src) \
	{ \
		(des)[0] = (src)[0]; \
		(des)[1] = (src)[1]; \
		(des)[2] = (src)[2]; \
		(des)[3] = (src)[3]; \
	}

#define MUL_MATRIX_ROW( k, des) \
	{ \
		(des)[0] = double((des)[0]) * k; \
		(des)[1] = double((des)[1]) * k; \
		(des)[2] = double((des)[2]) * k; \
		(des)[3] = double((des)[3]) * k; \
	}

#define ADD_MATRIX_ROW( des, src) \
	{ \
		(des)[0] = double((des)[0]) + double((src)[0]); \
		(des)[1] = double((des)[1]) + double((src)[1]); \
		(des)[2] = double((des)[2]) + double((src)[2]); \
		(des)[3] = double((des)[3]) + double((src)[3]); \
	}

#define ZERO_COLUMN( des, src, col, k) \
	{ \
		rgFloat macro_srcCpy[4]; \
		k = double((des)[col]) * -1.0; \
		MOV_MATRIX_ROW( macro_srcCpy, src); \
		MUL_MATRIX_ROW( k, macro_srcCpy); \
		ADD_MATRIX_ROW( des, macro_srcCpy); \
	}


bool rgMatrix::invert()
{
	/* Backup current matrix incase the inversion fails. */
	rgMatrix bak = *this;

	rgMatrix inv;
	if( !makeLeadingOne( 0, 0, inv)) { *this = bak; return false;}
	if( !makeLeadingOne( 1, 1, inv)) { *this = bak; return false;}
	if( !makeLeadingOne( 2, 2, inv)) { *this = bak; return false;}
	if( !makeLeadingOne( 3, 3, inv)) { *this = bak; return false;}

	/* The matrix has been inverted. */
	*this = inv;
	return true;
}

bool rgMatrix::rowElementLeads( rgFloat *d, int col)
{
	/* Unfortunatley, rowElementLeads requires perfect precision since it
	 * performs a critical test for 0.0f.
	 *
	 * TODO: The good news is that precisionGaurd is not neccasary because
	 *	it is possible to know when a matrix element "should" be zero.
	 *	Basically matrix inversion works by setting a leading one and
	 *	then zeroing out all other rows. Therefore the caller should
	 *	force all values which should be 0.0f to 0.0f. This way,
	 *	rowElementLeads will have no problem detecting zeros.
	 */
	precisionGaurd();	/* Make valuse which should be 0.0f, equal
				 * 0.0f.
				 * TODO: This is the caller's job.
				 */

	/* Requested element must be non zero. */
	if( 0.0f == d[col]) return false;

	/* All elements infront of the requested element must be zero. */
	switch( col) {
	case 3: if( d[2] != 0.0f) return false;
	case 2: if( d[1] != 0.0f) return false;
	case 1: if( d[0] != 0.0f) return false;
	}

	return true;
}

bool rgMatrix::makeLeadingOne( const int tarRow, const int tarCol,
	rgMatrix &inv)
{
	rgFloat *desRow;
	rgFloat *srcRow;
	rgFloat *desI;
	rgFloat *srcI;

	desRow = &d[ tarRow * 4];
	desI = &inv.d[ tarRow * 4];	

	if( rowElementLeads( d, tarCol)) {
		srcRow = &d[0]; srcI = &inv.d[0];
	} else if( rowElementLeads( d + 4, tarCol)) {
		srcRow = &d[4]; srcI = &inv.d[4];
	} else if( rowElementLeads( d + 8, tarCol)) {
		srcRow = &d[8]; srcI = &inv.d[8];
	} else if( rowElementLeads( d + 12, tarCol)) {
		srcRow = &d[12]; srcI = &inv.d[12];
	} else {
		/* Cannot make a leading one in target column because there is
		 * no possible linear combination of rows in the matrix to do
		 * it.
		 */
		return false;
	} 

	/* Make a leading one in target column */
	{
		const double k = 1.0f/srcRow[ tarCol];
		MUL_MATRIX_ROW( k, srcRow);
		MUL_MATRIX_ROW( k, srcI);
	}

	/* srcRow now points to a row that has the leading one. The caller
	 * however, requested the leading one to be in desRow. Therefore, swap
	 * srcRow with desRow (unless srcRow is equal to desRow, in which case
	 * do nothing).
	 */
	rgFloat temp[4];
	if( srcRow != desRow) {
		/* Swap the source and destination rows to make leading one. */
		MOV_MATRIX_ROW( temp, desRow);
		MOV_MATRIX_ROW( desRow, srcRow);
		MOV_MATRIX_ROW( srcRow, temp);

		/* Do the same thing to the identity matrix */
		MOV_MATRIX_ROW( temp, desI);
		MOV_MATRIX_ROW( desI, srcI);
		MOV_MATRIX_ROW( srcI, temp);
	}

	/* Zero out the entires in the target column of all the rows (except
	 * the row that contains the leading one).
	 */
	double scalar;
	if( desRow != &d[0]) {
		ZERO_COLUMN( &d[0], desRow, tarCol, scalar);
		MOV_MATRIX_ROW( temp, desI);
		MUL_MATRIX_ROW( scalar, temp);
		ADD_MATRIX_ROW( &inv.d[0], temp);
	}

	if( desRow != &d[4]) {
		ZERO_COLUMN( &d[4], desRow, tarCol, scalar);
		MOV_MATRIX_ROW( temp, desI);
		MUL_MATRIX_ROW( scalar, temp);
		ADD_MATRIX_ROW( &inv.d[4], temp);	
	}

	if( desRow != &d[8]) {
		ZERO_COLUMN( &d[8], desRow, tarCol, scalar);
		MOV_MATRIX_ROW( temp, desI);
		MUL_MATRIX_ROW( scalar, temp);
		ADD_MATRIX_ROW( &inv.d[8], temp);	
	}

	if( desRow != &d[12]) {
		ZERO_COLUMN( &d[12], desRow, tarCol, scalar);
		MOV_MATRIX_ROW( temp, desI);
		MUL_MATRIX_ROW( scalar, temp);
		ADD_MATRIX_ROW( &inv.d[12], temp);	
	}

	return true;
}

void rgMatrix::precisionGaurd()
{
	if( fabs(d[0]) < RG_MAT_MIN_FLOAT_VALUE) d[0] = 0.0f;
	if( fabs(d[1]) < RG_MAT_MIN_FLOAT_VALUE) d[1] = 0.0f;
	if( fabs(d[2]) < RG_MAT_MIN_FLOAT_VALUE) d[2] = 0.0f;
	if( fabs(d[3]) < RG_MAT_MIN_FLOAT_VALUE) d[3] = 0.0f;
	if( fabs(d[4]) < RG_MAT_MIN_FLOAT_VALUE) d[4] = 0.0f;
	if( fabs(d[5]) < RG_MAT_MIN_FLOAT_VALUE) d[5] = 0.0f;
	if( fabs(d[6]) < RG_MAT_MIN_FLOAT_VALUE) d[6] = 0.0f;
	if( fabs(d[7]) < RG_MAT_MIN_FLOAT_VALUE) d[7] = 0.0f;
	if( fabs(d[8]) < RG_MAT_MIN_FLOAT_VALUE) d[8] = 0.0f;
	if( fabs(d[9]) < RG_MAT_MIN_FLOAT_VALUE) d[9] = 0.0f;
	if( fabs(d[10]) < RG_MAT_MIN_FLOAT_VALUE) d[10] = 0.0f;
	if( fabs(d[11]) < RG_MAT_MIN_FLOAT_VALUE) d[11] = 0.0f;
	if( fabs(d[12]) < RG_MAT_MIN_FLOAT_VALUE) d[12] = 0.0f;
	if( fabs(d[13]) < RG_MAT_MIN_FLOAT_VALUE) d[13] = 0.0f;
	if( fabs(d[14]) < RG_MAT_MIN_FLOAT_VALUE) d[14] = 0.0f;
	if( fabs(d[15]) < RG_MAT_MIN_FLOAT_VALUE) d[15] = 0.0f;
}

rgFloat rgDotAngle( const rgVector a, const rgVector b)
{
	// ArcCos [ (A dot B) / (||A|| * ||B||) ]
	const double d = (a.x*b.x + a.y*b.y + a.z*b.z) /
		(sqrt(a.x*a.x + a.y*a.y + a.z*a.z) *
			sqrt(b.x*b.x + b.y*b.y + b.z*b.z));
	// Floating point is prone to errors, must check for invalid values.
	// NOTE: The dot product calculation above should never be more than
	//	1.0, nor less than -1.0.
	if( d < (-1.0)) return RG_180_DEG_IN_RAD;
	if( d > 1.0) return 0;
	return rgFloat( acos( d));
}

rgFloat rgAngle360( const rgVector origin, rgVector p)
{
	//          90
	//          |
	//          Y 
	//          | 
	// 180 -----+---X-> 0 degrees
	//          |
	//          |
	//         270

	rgVector v = (p - origin).normalize();
	float ang = v.dot( rgVector( 1, 0, 0));
	if( ang < (-1.0f)) ang = -1.0f;
	else if( ang > 1.0f) ang = 1.0f;
	if( v.dot( rgVector( 0, 1, 0)) >= 0) return acos( ang);
	else return RG_360_DEG_IN_RAD - acos( ang);
}

#ifdef USE_THIS_OLD_ANGLE_360_CODE
rgFloat rgAngle360( rgVector origin, rgVector p)
{
	/* Determine the quadrant that "p" is in.
	 *   2 | 1
	 *     | 
	 * ----+---- 0 degrees
	 *     |
	 *   3 | 4
	 */
	rgVector v = p - origin;

	rgFloat ang;
	if( (v.x >= 0.0f)) {
	/* In quadrant 1, or quadrant 4 */
		if( (v.y >= 0.0f)) {
		/* In quadrant 1 */
			ang = rgDotAngle( v, rgVector( 1.0f, 0.0f, 0.0f)); 
		} else {
		/* In quadrant 4 */
			ang =	RG_270_DEG_IN_RAD +
				rgDotAngle( v, rgVector( 0.0f, -1.0f, 0.0f)); 
		}
	} else {
	/* In quadrant 2, or quadrant 3 */
		if( (v.y >= 0.0f)) {
		/* In quadrant 2 */
			ang =	RG_90_DEG_IN_RAD +
				rgDotAngle( v, rgVector( 0.0f, 1.0f, 0.0f)); 
		} else {
		/* In quadrant 3 */
			ang =	RG_180_DEG_IN_RAD +
				rgDotAngle( v, rgVector( -1.0f, 0.0f, 0.0f)); 
		}
	}

	return ang;
}
#endif


int sphereRayIntersection( rgVector sCenter, rgFloat sRadius, rgVector rp,
        rgVector rd, rgVector &cp1, rgVector &cp2)
{
        const rgFloat a = rd.dot( rd);
        if( 0.0f == a) return 0;

        const rgFloat b = 2.0f * rd.dot( rp - sCenter);
        const rgFloat c = (rp - sCenter).dot( rp - sCenter) - sRadius * sRadius;
        rgFloat k = (b*b) - (4.0f * a * c);
        if( k < 0.0f) return 0;
	k = sqrt(k);

        const rgFloat t1 = (-b - k) / (2.0f * a);
        const rgFloat t2 = (-b + k) / (2.0f * a);

	if( t1 <= 0.0f) {
		if( t2 <= 0.0f) return 0;
	        else {
			cp1 = (t2 * rd) + rp;
			return 1;
		}
	} else {
	        cp1 = (t1 * rd) + rp;
		if( t2 <= 0.0f) return 1;
	        else cp2 = (t2 * rd) + rp;
	}

        if( 0.0f == k) return 1;

        return 2;
}

/*
rgVector rgProjectAlong( rgVector v, rgVector u)
{
	rgFloat lengthSquared = u.length();
	lengthSquared *= lengthSquared;
	if( 0.0f == lengthSquared) return rgVector();

	const rgFloat k = u.dot( v) / lengthSquared;

	return k * u;
}
*/

/*
rgVector rgReflectAbout( rgVector v, rgVector n)
{
	return v + (2.0f * (rgProjectAlong( v, n) - v));
}
*/

rgFloat rgPointPlaneDis( rgVector &pp, rgVector &pn, rgVector &p)
{
	return fabs( (p - pp).dot( pn));
}

bool rgRayPlaneIntersect( rgVector pp, rgVector pn, rgVector rp, rgVector rd,
	rgVector &ip)
{
	rgFloat t;
	rgLinePlaneIntersectScalar( pp, pn, rp, rd, t);
	/* FIXME: This commented out code is most likely not needed.
	if( !rgLinePlaneIntersectScalar( pp, pn, rp, rd, t)) {

		// Check if ray point is on the plane.
		if( 0.0f == pn.dot( rp - pp)) {
			ip = rp;
			return true;
		}
		return false;
	}
	*/

	if( t >= 0.0f) {
		ip = rp + (t * rd);
		return true;
	}

	return false;
}

bool rgLineSegPlaneIntersect( rgVector pp, rgVector pn, rgVector a,
	rgVector b, rgVector &ip)
{
	rgFloat t;
	return rgLineSegPlaneIntersect( pp, pn, a, b, ip, t);
}

bool rgLineSegPlaneIntersect( rgVector pp, rgVector pn, rgVector a,
	rgVector b, rgVector &ip, rgFloat &t)
{
	rgVector ba = b - a;
	if( !rgLinePlaneIntersectScalar( pp, pn, a, ba, t)) {

		/* Check if ray point is on the plane. */
		if( 0.0f == pn.dot( a - pp)) {
			ip = a;
			t = 0.0f;
			return true;
		}
		return false;
	}

	if( t >= 0.0f && t <= 1.0f) {
		ip = a + (t * ba);
		return true;
	}

	return false;
}

bool rgLinePlaneIntersect( rgVector pp, rgVector pn, rgVector lp, rgVector ld,
	rgVector &ip)
{
	rgFloat t;
	if( !rgLinePlaneIntersectScalar( pp, pn, lp, ld, t)) {

		/* Check if ray point is on the plane. */
		if( 0.0f == pn.dot( lp - pp)) {
			ip = lp;
			return true;
		}
		return false;
	}

	ip = lp + (t * ld);
	return true;
}

inline bool rgLinePlaneIntersectScalar( rgVector &pp, rgVector &pn,
	rgVector &lp, rgVector &ld, rgFloat &t)
{
	const rgFloat k = pn.dot( ld);
	if( 0.0f == k) return false;

	t = (pn.dot( pp) - pn.dot( lp)) / k;

	return true;
}

/*
bool rgPlanarPointInConvexFace( const rgVector *f, const int count,
	const rgVector p)
{
	if( count < 3) return false;
	const rgVector bn( (f[1] - f[0]).cross( p - f[0]));
	
	for( int i = 1; i < (count - 1); i++) {
		const rgVector n( ( f[i+1] - f[i]).cross( p - f[i]));
		if( bn.dot( n) <= 0.0f) return false;
	}

	const rgVector n( ( f[0] - f[count - 1]).cross( p - f[count - 1]));
	if( bn.dot( n) <= 0.0f) return false;

	return true;
}
*/

bool rgRayTriangleIntersect( rgVector a, rgVector b, rgVector c, rgVector n,
	rgVector rp, rgVector rd, rgVector &ip)
{
	if( !rgRayPlaneIntersect( a, n, rp, rd, ip)) return false;
	if( !rgPlanarPointInTriangle( a, b, c, ip)) return false;
	return true;
}


bool rgLineLineIntersect( rgVector a, rgVector b, rgVector c,
	rgVector d, rgVector &p)
{

	/* Calculate a "normal-like" vector. */
	rgVector u( b - a);
	rgVector v( c - a);
	rgVector n( v - rgProjectAlong( v, u));

	const rgFloat t = v.dot( n);
	rgVector r( (d - c).normalize());
	if( 0.0f == r.dot( n)) {
		// Lines are parallel. There is an intersection only if
		// "t" is zero.
		if( t != 0.0f) return false;
		p = a;
		return true;
	}

	// Lines are not parallel, so there must be an intersection point.
	p = c + (t * r);
	return true;
}

bool rgRayLineSegIntersect( rgVector rp, rgVector rd, rgVector c,
	rgVector d, rgVector &p)
{
	rgFloat t, k;
	return rgRayLineSegIntersect( rp, rd, c, d, p, t, k);
}

bool rgRayLineSegIntersect( rgVector rp, rgVector rd, rgVector c,
	rgVector d, rgVector &p, rgFloat &t, rgFloat &k)
{
	rgVector crp = c - rp;

	/* Calculate a "normal-like" vector. */
	rgVector n = crp - rgProjectAlong( crp, rd);

	const rgFloat crp_dot_n = crp.dot( n);
	if( 0.0f == crp_dot_n) {
		t = 0.0f;
		p = c;

	} else {
		const rgFloat dc_dot_n = (d - c).dot( n);
		if( 0.0f == dc_dot_n) return false;

		t = (rp.dot( n) - c.dot( n)) / dc_dot_n;
		if( (t < 0.0f) || (t > 1.0f)) return false;

		p = c + ((d - c) * t);
	}

	rgFloat const rd_dot_rd = rd.dot( rd);
	if( 0.0f == rd_dot_rd) {
		if( p == rp) {
			k = 0.0f;
			return true;
		}
		return false;
	}

	k = (p - rp).dot( rd) / rd_dot_rd;
	if( k < 0.0f) return false;

	return true;
}

bool rgLineLineSegIntersect( rgVector a, rgVector b, rgVector c,
	rgVector d, rgVector &p)
{
	rgVector ba = b - a;
	rgVector ca = c - a;

	/* Calculate a "normal-like" vector. */
	rgVector n = ca - rgProjectAlong( ca, ba);

	const rgFloat ca_dot_n = ca.dot( n);
	if( 0.0f == ca_dot_n) {
		p = c;
		return true;
	}
	
	const rgFloat dc_dot_n = (d - c).dot( n);
	if( 0.0f == dc_dot_n) return false;

	const rgFloat t = (a.dot( n) - c.dot( n)) / dc_dot_n;
	if( (t < 0.0f) || (t > 1.0f)) return false;

	p = c + ((d - c) * t);

	return true;
}

/*
bool rgLineSegLineSegIntersect( rgVector a, rgVector b, rgVector c,
	rgVector d, rgVector &p)
{
	rgFloat t, k;
	return rgLineSegLineSegIntersect( a, b, c, d, p, t, k);
}
*/

/*
bool rgLineSegLineSegIntersect( rgVector a, rgVector b, rgVector c,
	rgVector d, rgVector &p, rgFloat &t, rgFloat &k)
{
	const rgVector ba( b - a);
	const rgVector ca( c - a);

	// Calculate a "normal-like" vector.
	const rgVector n( ca - rgProjectAlong( ca, ba));

	if( 0.0f == ca.dot( n)) {
		t = 0.0f;
		p = c;

	} else {	
		const rgFloat dc_dot_n = (d - c).dot( n);
		if( 0.0f == dc_dot_n) return false;

		t = (a.dot( n) - c.dot( n)) / dc_dot_n;
		if( (t < 0.0f) || (t > 1.0f)) return false;

		p = c + ((d - c) * t);
	}

	const rgFloat ba_dot_ba = ba.dot( ba);
	if( 0.0f == ba_dot_ba) {
		if( p == a) {
			k = 0.0f;
			return true;
		}
		return false;
	}

	k = (p - a).dot( ba) / ba_dot_ba;
	if( (k < 0.0f) || (k > 1.0f)) return false;

	return true;
}
*/

int rgCircleRayIntersect( rgVector cp, const rgFloat r, rgVector rp,
	rgVector rd, rgVector &ip1, rgVector &ip2)
{
	rgVector v = rgProjectAlong( (cp - rp), rd);
	rd = rd.normalize();

	const rgFloat d = (v - (cp - rp)).length();
	const rgFloat t = (r * r) - (d * d);
	if( t < 0.0f) return 0;
	if( t == 0.0f) {
		ip1 = v;
		return 1;
	}

	ip1 = (rgFloat( sqrt( t)) * rd);
	ip2 = v + ip1;
	ip1 = v - ip1;
	return 2;
}

rgVector rgMapToBasis( rgVector p, rgBasis &ob, rgBasis &nb)
{
	return	(p.dot( ob.side) * nb.side) +	/* "X" axis */
		(p.dot( ob.up) * nb.up) +	/* "Y" axis */
		(p.dot( ob.eye) * nb.eye);	/* "Z" axis */
}


