/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef RGTIME_H
#define RGTIME_H


class rgTimer
{
public:

	rgTimer();
	rgTimer( const double sec);

	// Description: Returns the number of elapsed micro seconds since
	// 	the last call to elapsed or the since rgTimer was created.
	double elapsed();
	double elapsed_mili_sec() { return elapsed() * 1000.0f;}

	double elapsed_scalar() { return elapsed() / double( time_unit);}
	double get_time_unit() const { return time_unit;}

	bool time_unit_passed() const;

	// Description: Resets the timer.
	void reset();
	
	void set_time_unit( const double sec) {
		if( sec <= 0.0) return;
		time_unit = sec;
	}

private:
	double sec_prev;
	double time_unit;

	void init();
};

class rgFPSCounter
{
public:
        int count_reset;
	double fps;

	rgFPSCounter();
	rgFPSCounter( double sec);

	void inc_frame();

	void set_time_unit( double sec);
	bool time_unit_passed();

private:
	double fcount;
	rgTimer timer;
};

// Description: Delays for "approximatley" the specified number of seconds.
//	"rgTime_delay" is not interrupt based and so is not perfectly accurate.
//	"rgTime_delay" should be accurate to within a few micro seconds. Also,
//	"rgTime_delay" returns the percentage of time delayed to allow the
//	caller to know of the timing accuracy.
//
//	NOTE: "rgTime_delay" is subject to function call overhead (as well as
//		operating system processing overhead) so one should expect
//		these delays.
//
// Return: 1.0 is returned if rgTime_delay delayed for exactly "sec" seconds.
//	Otherwise, a value greater than 1.0 is returned if "rgTime_delay"
//	delayed for longer than "sec" seconds.
double rgTime_delay( const double sec);

#endif
